/* Voronoi Diagram plug-in
 * Copyright (C) 1999 Joachim Wagner
 *
 * The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Voronoi Diagram plug-in version 0.01, Dezember 1999
 *
 * This plug-in generates ...
 *
 * You can contact me at <jowagner@uos.de>.
 * Comments and improvements for this code are welcome.
 *
 * The overall plug-in structure is based on the Solid Noise
 * plug-in, which is Copyright (C) 1997, 1998 Marcelo de 
 * Gomensoro Malheiros and which itself is based on the Whirl 
 * plug-in, which is Copyright (C) 1997 Federico Mena Quintero.
 *
 * @?(GIMP)         = <plug-in voronoi "Filters/Render/voronoi">
 * @?(GIMP_DEP)     = <voronoi.c>
 * @?(GIMP_OBJ)     = <voronoi.o>
 * @?(GIMP_LIB)     = <c m>
 * @?(GIMP_AUTHOR)  = <Joachim Wagner>
 * @?(GIMP_EMAIL)   = <jowagner@uos.de>
 * @?(GIMP_DESC)    = <Voronoi Diagram>
 * @?(GIMP_VERSION) = <1.00>
 * @?(GIMP_URL)     = <http://www-lehre.informatik.uni-osnabrueck.de/~jwagner/>
 */

/*
 *
 * Version 0.02:
 *
 *  added grid
 *
 * Version 0.01:
 *
 *  Initial release.
 *  
 *  2do:
 *    - ~/gimp/rand-noted.c lesen
 *    - Eingabe: Nullen streichen nach .0
 *    - Tilebenutzung optimieren
 *    - 
 */

#include <time.h>
#include <math.h>
#include <stdlib.h>

#include "gtk/gtk.h"
#include "libgimp/gimp.h"


/*---- Defines ----*/

#define ENTRY_WIDTH 40    

#define EPSILON (pow(2,-120))

/*---- Typedefs ----*/

typedef struct voronoiValues {
  gdouble p;
  gdouble jitter;
  gdouble diameter;
  gdouble linesickness;
  gdouble nod_threshold;
  gdouble grid_distance;
  gint    samples;
  gint    seed;
  gint    showpoints;
  gint    showlines;
  gint    numpoints;
  gint    use_p_norm;
  gint    use_max_norm;
  gint    use_nod_norm;
  gint    color_class;
  gint    color_norm;
  gint    use_grid;
} voronoiValues;

typedef struct voronoiDialogData {
  char *label;
  void *pdata;
  void (*func)(GtkWidget *widget, gpointer data);
  void (*sprint)(char *buffer, void *pdata);
} voronoiDialogData;

typedef struct {
  gint run;
} voronoiInterface;

typedef struct Vector2d{
  double x, y;
} Vector2d;


/*---- Prototypes ----*/


static void query (void);
static void run (char *name, int nparams, GParam  *param,
                 int *nreturn_vals, GParam **return_vals);

static gint voronoi (GDrawable *drawable);
static gint voronoi_init (GDrawable *drawable, gint width, gint height);
static void voronoi_deinit (void);

static int plain_class (double x, double y, double *pd);
static int near_point (double x, double y);
static double voronoi_pixel (double x, double y); 
static double norm (double x, double y);
static void permutpoints (void);

static gint voronoi_dialog (void);
static void dialog_close_callback (GtkWidget *widget, gpointer data);
static void dialog_toggle_update (GtkWidget *widget, gpointer data);
static void dialog_float_entry_callback (GtkWidget *widget, gpointer data);
static void dialog_int_entry_callback (GtkWidget *widget, gpointer data);
static void dialog_sprint_float(char *buffer, void *pdata);
static void dialog_sprint_int(char *buffer, void *pdata);
static void dialog_ok_callback (GtkWidget *widget, gpointer data);


/*---- Variables ----*/

GPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc */
  NULL,  /* quit_proc */
  query, /* query_proc */
  run,   /* run_proc */
};

static voronoiValues snvals = {
    2.0,     /* p (p-norm)        */
   25.0,     /* jitter (samples)  */
  400.0,     /* diameter of points (if shown) */
  200.0,     /* linesickness      */
   20.0,     /* threshold of nod-norm */
   50.0,     /* distance of grid-lines */
  5,         /* samples per pixel */
  27,        /* random seed       */
  1,         /* show points       */
  1,         /* show lines        */
  15,        /* Number of Points  */
  1,         /* 1 = use p-norm    */
  0,         /* 1 = use max-norm  */
  0,         /* 1 = use number-of-differences-norm  */
  1,         /* 1 = use color to show class */
  0,         /* 1 = use color to show norm */
  1          /* 1 = use grid for points */
};

static voronoiInterface snint = {
  FALSE /* run */
};

static Vector2d *point;

static const double pi = 3.14159265358979323846;

static Vector2d *sample;
static gint sm1, sm2;
static gdouble jitter;

static gdouble norm_max;
static gint *permutpoint;

static gint sel_x1, sel_y1,
            sel_x2, sel_y2,
            sel_width, sel_height;
static gint chns, has_alpha;

              
/*---- Functions ----*/

MAIN()


static void
query (void)
{
  static GParamDef args[] =
  {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_IMAGE, "image", "Input image" },
    { PARAM_DRAWABLE, "drawable", "Input drawable" },

    { PARAM_FLOAT, "p",          "p Parameter of L-Norm" },
    { PARAM_FLOAT, "jitter",     "jitter in percent" },
    { PARAM_FLOAT, "diameter",   "diameter of points in percent (if shown)" },
    { PARAM_FLOAT, "nod_threshold","thresold of nod-norm" },
    { PARAM_FLOAT, "grid_distance","distance of grid-lines" },
    { PARAM_INT32, "samples",      "Samples per pixel" },
    { PARAM_INT32, "seed",         "Random seed" },
    { PARAM_INT32, "showpoints",   "show point = 1, do not = 0" },
    { PARAM_INT32, "showlines",    "show lines = 1, do not = 0" },
    { PARAM_INT32, "numpoints",    "number of points" },
    { PARAM_INT32, "use_p_norm",   "1 = use p-norm" },
    { PARAM_INT32, "use_max_norm", "1 = use max-norm" },
    { PARAM_INT32, "use_nod_norm", "1 = use number-of-differences-norm" },
    { PARAM_INT32, "color_class",  "1 = use color to show class" },
    { PARAM_INT32, "color_norm",   "1 = use color to show norm" },
    { PARAM_INT32, "use_grid",     "1 = use grid for points" }
  };
  static GParamDef *return_vals = ((void *)0) ;
  static gint nargs = sizeof (args) / sizeof (args[0]);
  static gint nreturn_vals = 0;

  gimp_install_procedure ("plug_in_voronoi",
                          "Creates a voronoi diagram",
                          "Generates a Voronoi diagram",
                          "Joachim Wagner",
                          "Joachim Wagner",
                          "Dec 1999, v0.02",
			  "<Image>/Filters/Render/voronoi",
			  "RGB*, GRAY*",
			  PROC_PLUG_IN,
			  nargs,
                          nreturn_vals,
			  args,
                          return_vals);
}


static void
run (char *name, int nparams, GParam *param, int *nreturn_vals,
     GParam **return_vals)
{
  static GParam values[1];
  
  GDrawable *drawable;
  GRunModeType run_mode;
  GStatusType status;
  
  status = STATUS_SUCCESS;
  run_mode = param[0].data.d_int32;

  values[0].type = PARAM_STATUS;
  values[0].data.d_status = status;

  *nreturn_vals = 1;
  *return_vals  = values;
  
  /*  Get the specified drawable  */
  drawable = gimp_drawable_get (param[2].data.d_drawable);

  /*  See how we will run  */
  switch (run_mode) {
  case RUN_INTERACTIVE:
    /*  Possibly retrieve data  */
    gimp_get_data("plug_in_voronoi", &snvals);

    /*  Get information from the dialog  */
    if (!voronoi_dialog ())
      return;

    break;

  case RUN_NONINTERACTIVE:
    /*  Test number of arguments  */
    if (nparams == 20) {
      snvals.p            = param[ 3].data.d_float;
      snvals.jitter       = param[ 4].data.d_float;
      snvals.diameter     = param[ 5].data.d_float;
      snvals.linesickness = param[ 6].data.d_float;
      snvals.nod_threshold= param[ 7].data.d_float;
      snvals.grid_distance= param[ 8].data.d_float;
      snvals.samples      = param[ 9].data.d_int32;
      snvals.seed         = param[10].data.d_int32;
      snvals.showpoints   = param[11].data.d_int32;
      snvals.showlines    = param[12].data.d_int32;
      snvals.numpoints    = param[13].data.d_int32;
      snvals.use_p_norm   = param[14].data.d_int32;
      snvals.use_max_norm = param[15].data.d_int32;
      snvals.use_nod_norm = param[16].data.d_int32;
      snvals.color_class  = param[17].data.d_int32;
      snvals.color_norm   = param[18].data.d_int32;
      snvals.use_grid     = param[19].data.d_int32;
    } 
    else
      status = STATUS_CALLING_ERROR;
    break;

  case RUN_WITH_LAST_VALS:
    /*  Possibly retrieve data  */
    snvals.p = -1.0;
    gimp_get_data("plug_in_voronoi", &snvals);
    if (snvals.p < 0.0) 
      return;
    break;
    
  default:
    break;
  }
  
  /*  Create image */
  if ((status == STATUS_SUCCESS) && (gimp_drawable_color (drawable->id) ||
      gimp_drawable_gray (drawable->id)))
    {
      /*  Set the tile cache size  */
      gimp_tile_cache_ntiles(6 + (drawable->width + gimp_tile_width() - 1) 
                                  / gimp_tile_width());

      /*  Run!  */
      if (!voronoi (drawable))
        status = STATUS_EXECUTION_ERROR;

      /*  If run mode is interactive, flush displays  */
      if (run_mode != RUN_NONINTERACTIVE)
        gimp_displays_flush();

      /*  Store data  */
      if (run_mode==RUN_INTERACTIVE || run_mode==RUN_WITH_LAST_VALS)
        gimp_set_data("plug_in_voronoi", &snvals,
                      sizeof(voronoiValues));
    }
  else
    {
      /* gimp_message ("voronoi: cannot operate on indexed color images"); */
      status = STATUS_EXECUTION_ERROR;
    }  
  
  values[0].data.d_status = status;
  
  gimp_drawable_detach(drawable);
}


static gint
voronoi (GDrawable *drawable)
{
  GPixelRgn dest_rgn;
  gint i, row, col;
  gint progress, max_progress;
  gpointer pr;
  guchar *dest, *dest_row;
  guchar val;

  /*  Get selection area  */
  gimp_drawable_mask_bounds (drawable->id, &sel_x1, &sel_y1, &sel_x2, &sel_y2);
  sel_width = sel_x2 - sel_x1;
  sel_height = sel_y2 - sel_y1;

  /*  Initialization  */
  if (!voronoi_init (drawable, sel_width, sel_height))
    return 0;
  gimp_progress_init("voronoi (2/2)...");
  gimp_progress_update(0.0);
  progress = 0;
  max_progress = sel_width * sel_height;
  chns = gimp_drawable_bpp (drawable->id);
  has_alpha = 0;
  if (gimp_drawable_has_alpha (drawable->id)) {
    chns--;
    has_alpha = 1;
  }
  gimp_pixel_rgn_init (&dest_rgn, drawable, sel_x1, sel_y1, sel_width,
                       sel_height, TRUE, TRUE);

  /*  One, two, three, go!  */
  for (pr = gimp_pixel_rgns_register (1, &dest_rgn); pr != NULL;
       pr = gimp_pixel_rgns_process (pr))
    {
      dest_row = dest_rgn.data;

      for (row = dest_rgn.y; row < (dest_rgn.y + dest_rgn.h); row++)
        {
          dest = dest_row;
          
          for (col = dest_rgn.x; col < (dest_rgn.x + dest_rgn.w); col++)
            {
              val = (guchar) floor (255.0 * voronoi_pixel(col - sel_x1 , 
                                                  row - sel_y1 ));
              for (i = 0; i < chns; i++)
                *dest++ = val;
              if (has_alpha)
                *dest++ = 255;
            }

          dest_row += dest_rgn.rowstride;;
        }

      /*  Update progress  */
      progress += dest_rgn.w * dest_rgn.h;
      gimp_progress_update ((double) progress / (double) max_progress);
    }

  /*  Update the drawable  */
  gimp_drawable_flush (drawable);
  gimp_drawable_merge_shadow (drawable->id, TRUE);
  gimp_drawable_update (drawable->id, sel_x1, sel_y1, sel_width, sel_height);
  voronoi_deinit();
  return 1;
}


static gint
voronoi_init (GDrawable *drawable, gint width, gint height)
{
  gdouble xp, yp, a, h, r, xpp, ypp;
  gint i, j, versuch, neuer_punkt;
  gint samples, swidth;

  gimp_progress_init("voronoi (1/2)...");

  /*  Force sane parameters  */
  if (snvals.diameter <= 0.0) {
    snvals.diameter=100; fprintf(stderr, "Warning: diameter set to 100%%\n"); }
  if ((snvals.jitter < 0.0) || (snvals.jitter >200.0)) {
    fprintf(stderr, "Warning: Jitter %f%% abnormal - not changed\n",
                    snvals.jitter); }
  if (snvals.samples < 0) {
    snvals.samples = 0; fprintf(stderr, "Warning: samples set to 0\n"); }
  if (snvals.samples > 1024) {
    fprintf(stderr, "Warning: samples %d abnormal - not changed\n",
                    snvals.samples); }
  if (snvals.numpoints < 1) {
    snvals.numpoints = 1; fprintf(stderr, "Warning: numpoints set to 1\n"); }
  if (snvals.numpoints > 1024) {
    fprintf(stderr, "Warning: numpoints %d abnormal - not changed\n",
                    snvals.numpoints); }

  /*  Define pseudo-random number generator seed  */

  srand (snvals.seed);

  /* setup points */

  point = g_new(Vector2d, snvals.numpoints);
  permutpoint = g_new(gint, snvals.numpoints);

  for (i=0; i<snvals.numpoints; i++) {
    versuch = 0;
    do {
      xp = sel_width * (rand()&1023)/1024.0;
      yp = sel_height * (rand()&1023)/1024.0;
      if (snvals.use_grid) {
        xp += snvals.grid_distance / 2.0;
        yp += snvals.grid_distance / 2.0;
        xp /= snvals.grid_distance;
        yp /= snvals.grid_distance;
        xp = floor(xp);
        yp = floor(yp);
        xp *= snvals.grid_distance;
        yp *= snvals.grid_distance;
      }
      neuer_punkt = 1;
      for (j=0; j<i; j++) {
        if ( (fabs(point[j].x - xp) < EPSILON) 
           &&(fabs(point[j].y - yp) < EPSILON) ) 
        {
           fprintf(stderr,"Punkt p[%3i] = (%lf, %lf) Konflikt zu "
                                "p[%3i] = (%lf, %lf)\n", 
                           i, xp, yp, j, point[j].x, point[j].y);
           neuer_punkt = 0;
           /* break; */
        }
      }
      versuch++;
    }while((!neuer_punkt) && (versuch < 100));
    if (!neuer_punkt) {
      char temp[80];
      fprintf(stderr, "Keinen Platz fuer Punkt p[%3i] nach %i "
                      "Versuchen gefunden.\n", i, versuch);
      for (j=0; j<i; j++) 
        fprintf(stderr,"p[%3i] = (%lf, %lf)\n", j, point[j].x, point[j].y);
      /* besser: Fehlermeldung in Fenster */
      sprintf(temp, "No place left for point %i.\n", i+1);
      gimp_message (temp); 
      return 0;
    }
    point[i].x = xp;
    point[i].y = yp;
    permutpoint[i] = i;
    if (!(~i & 4095))
      gimp_progress_update(0.10 + 0.85*((double)i/(double)snvals.numpoints));
  }
  permutpoints();

  /* sampling setup */

  sm2 = snvals.samples - 1;
  swidth = ceil(sqrt(snvals.samples));
  samples = swidth * swidth;
  samples = swidth * swidth;
  sm1 = samples - 1;

  if (!sm2) sm2 = 1;

  sample = g_new(Vector2d, samples);
  for (i=0; i<swidth; i++) {
    yp = (double)(i*2+1)/(double)(swidth*2);
    for (j=0; j<swidth; j++) {
      xp = (double)(j*2+1)/(double)(swidth*2);
      xpp = xp - 0.5;
      ypp = yp - 0.5;
      if (fabs(xpp)>EPSILON) {
        a = atan(fabs(ypp)/fabs(xpp));
        h = tan(a) * 0.5;
        r = sqrt(0.25+h*h);
        xpp = xpp * 0.5 / r;
        ypp = ypp * 0.5 / r;
      }
      xpp += 0.5;
      ypp += 0.5;
      sample[i*swidth+j].x = xpp;
      sample[i*swidth+j].y = ypp;
      gimp_progress_update(0.95 + 0.05*((double)(i*swidth+j)/(double)samples));
    }
  }
  jitter = snvals.jitter / (100.0 * swidth);

  norm_max = norm(width,height);

  /* debug */

  fprintf(stderr,"--- voronoi ---\n");
  fprintf(stderr,"p             %f\n", snvals.p);
  fprintf(stderr,"jitter        %f\n", snvals.jitter);
  fprintf(stderr,"diameter      %f\n", snvals.diameter);
  fprintf(stderr,"linesickness  %f\n", snvals.linesickness);
  fprintf(stderr,"nod-threshold %f\n", snvals.nod_threshold);
  fprintf(stderr,"grid-distance %f\n", snvals.grid_distance);
  fprintf(stderr,"samples       %d\n", snvals.samples);
  fprintf(stderr,"seed          %d\n", snvals.seed);
  fprintf(stderr,"showpoints    %d\n", snvals.showpoints);
  fprintf(stderr,"showlines     %d\n", snvals.showlines);
  fprintf(stderr,"use p-norm    %d\n", snvals.use_p_norm);
  fprintf(stderr,"use max-norm  %d\n", snvals.use_max_norm);
  fprintf(stderr,"use nod-norm  %d\n", snvals.use_nod_norm);
  fprintf(stderr,"color class   %d\n", snvals.color_class);
  fprintf(stderr,"color norm    %d\n", snvals.color_norm);
  fprintf(stderr,"use grid      %d\n", snvals.use_grid);
  for (i=0; i<swidth; i++) {
    for (j=0; j<swidth; j++) {
      xp = sample[i*swidth+j].x;
      yp = sample[i*swidth+j].y;
      fprintf(stderr, "[%2i][%2i] (%f,%f)\n", j, i, xp, yp);
    }
  }
  fprintf(stderr,"jitter_eff  %f\n", jitter); 

  /* init finished */

  gimp_progress_update(1.00);
  return 1;
}

static void
voronoi_deinit (void)
{
  g_free(point);
  g_free(sample);
}

static void
permutpoints (void)
{
  int i, j, h;
  for (i=0; i<snvals.numpoints; i++) {
    j = rand() % snvals.numpoints;
    h = permutpoint[i];
    permutpoint[i] = permutpoint[j];
    permutpoint[j] = h;
  }
}


static double p_norm (double x, double y)
{
  x = fabs(x);
  y = fabs(y);
  return (pow(pow(x,snvals.p)+pow(y,snvals.p),
          1.0 / snvals.p));
}

static double max_norm (double x, double y)
{
  x = fabs(x);
  y = fabs(y);
  return (x>y)?x:y;
}

static double nod_norm (double x, double y)
{
  x = fabs(x);
  y = fabs(y);
  if (x > snvals.nod_threshold) {
    if (y > snvals.nod_threshold) {
      return 2.0*snvals.nod_threshold;
    }else{
      return 1.0*snvals.nod_threshold;
    }
  }else{
    if (y > snvals.nod_threshold) {
      return 1.0*snvals.nod_threshold;
    }else{
      return 0.0;
    }
  }
}

static double norm (double x, double y)
{
  int i = 0;
  double d=0.0;
  if (snvals.use_p_norm)   { d += p_norm(x,y);   i++; }
  if (snvals.use_max_norm) { d += max_norm(x,y); i++; }
  if (snvals.use_nod_norm) { d += nod_norm(x,y); i++; }
  if (i) return d/i;
  else return -1.0;
}

static int
plain_class (double x, double y, double *pd)
{
  double xd, yd, dmin, d;
  int i, j, class;
  i = permutpoint[0]; 
  class = i;
  xd = point[i].x - x;
  yd = point[i].y - y;
  dmin = norm(xd,yd);
  for (j=1; j<snvals.numpoints; j++) {
    i = permutpoint[j]; 
    xd = point[i].x - x;
    yd = point[i].y - y;
    if ((d=norm(xd,yd))<dmin) {
      dmin = d;
      *pd = d;
      class = i;
    }
  }
  permutpoints();
  return class;
}

static int
near_point (double x, double y)
{
  double xd, yd;
  int i, p;
  p = 0;
  for (i=0; i<snvals.numpoints; i++) {
    xd = point[i].x - x;
    yd = point[i].y - y;
    if (sqrt(xd*xd+yd*yd)<(snvals.diameter/100.0)) {
      p = 1;
      break;
    }
  }
  return p;
}


static double voronoi_pixel (double x, double y)
{
  int i, class;
  double sum;
  double xp, yp, c, c2, dummy;
  sum = 0.0;
  c = 1.0;
  class = -1;
  for (i = 0; i < snvals.samples; i++) {
    xp = x + snvals.linesickness/100.0
             *(sample[i*sm1/sm2].x + jitter*((rand()&1023) - 511.5)/1024.0);
    yp = y + snvals.linesickness/100.0
             *(sample[i*sm1/sm2].y + jitter*((rand()&1023) - 511.5)/1024.0);
    if (class < 0) {
      class = plain_class (xp, yp, &c2);
    }else{
      if (snvals.showlines && (class != plain_class (xp,yp, &dummy))) {
        c = 0.0;
        break;
      }
    }
  }
  if (c>0.5) {
    if (snvals.color_class) {
      c /= pow(2, 0.05 + 0.90 * class/snvals.numpoints);
    }
    if (snvals.color_norm) {
      c /= pow(2, 0.05 + 0.90 * c2/norm_max);
    }
  }
  if (snvals.showpoints && near_point(x,y)) {
    c = 0.15 + 0.10 * c;
  }
  return c;
}


static gint voronoi_dialog (void)
{
  GtkWidget *dlg;
  GtkWidget *toggle;
  GtkWidget *table;
  GtkWidget *button;
  GtkWidget *label;
  GtkWidget *entry;
  gchar **argv, **temp_ptr;
  gchar buffer[32];
  gint  argc, nums, i;
  voronoiDialogData num[] = {
     { "P",            &snvals.p,        dialog_float_entry_callback,
                                         dialog_sprint_float},
     { "nod-threshold", &snvals.nod_threshold, dialog_float_entry_callback,
                                         dialog_sprint_float},
     { "jitter (%)",   &snvals.jitter,   dialog_float_entry_callback,
                                         dialog_sprint_float},
     { "diameter (%)", &snvals.diameter, dialog_float_entry_callback,
                                         dialog_sprint_float},
     { "linesickness (%)", &snvals.linesickness, dialog_float_entry_callback,
                                         dialog_sprint_float},
     { "samples",      &snvals.samples,  dialog_int_entry_callback,
                                         dialog_sprint_int},
     { "seed",         &snvals.seed,     dialog_int_entry_callback,
                                         dialog_sprint_int},
     { "# of points",  &snvals.numpoints,dialog_int_entry_callback,
                                         dialog_sprint_int},
     { "grid distance", &snvals.grid_distance, dialog_float_entry_callback,
                                         dialog_sprint_float},
  };                      
  nums = sizeof(num)/sizeof(num[0]);

  /*  Set args  */
  argc = 1;
  argv = temp_ptr = g_new (gchar *, 1);  /* remember ptr 4 free */
  argv[0] = g_strdup ("voronoi");
  gtk_init (&argc, &argv);
  gtk_rc_parse (gimp_gtkrc ());

  /*  Dialog initialization  */
  dlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dlg), "voronoi");
  gtk_window_position (GTK_WINDOW (dlg), GTK_WIN_POS_MOUSE);
  gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dlg)->action_area), 2);
  gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
                      (GtkSignalFunc) dialog_close_callback, NULL);

  /*  Table  */
  table = gtk_table_new (nums, 3, FALSE);
  gtk_container_border_width (GTK_CONTAINER (table), 2);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), table, TRUE, TRUE, 0);

  /* Numerical Entries */

  for (i=0; i<nums; i++) {
    label = gtk_label_new (num[i].label);
    gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
    gtk_table_attach (GTK_TABLE (table), label, 0, 1, i, i+1,
                      GTK_FILL, GTK_FILL, 1, 0);
    gtk_widget_show (label);
  
    entry = gtk_entry_new ();
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, i, i+1,
                      GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_usize (entry, ENTRY_WIDTH, 0);
    num[i].sprint(buffer, num[i].pdata);
    gtk_entry_set_text (GTK_ENTRY (entry), buffer);
    gtk_signal_connect (GTK_OBJECT (entry), "changed",
                        (GtkSignalFunc) num[i].func, 
                        num[i].pdata);
    gtk_widget_show (entry);
  }

  /*  Check button #1  */
  toggle = gtk_check_button_new_with_label ("use p-norm");
  gtk_table_attach (GTK_TABLE (table), toggle, 2, 3, 0, 1,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
                               snvals.use_p_norm);
  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
                      (GtkSignalFunc) dialog_toggle_update,
                      &snvals.use_p_norm);
  gtk_widget_show (toggle);

  /*  Check button #2  */
  toggle = gtk_check_button_new_with_label ("use number-of-differences-norm");
  gtk_table_attach (GTK_TABLE (table), toggle, 2, 3, 1, 2,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
                               snvals.use_nod_norm);
  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
                      (GtkSignalFunc) dialog_toggle_update,
                      &snvals.use_nod_norm);
  gtk_widget_show (toggle);

  /*  Check button #3  */
  toggle = gtk_check_button_new_with_label ("use max-norm");
  gtk_table_attach (GTK_TABLE (table), toggle, 2, 3, 2, 3,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
                               snvals.use_max_norm);
  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
                      (GtkSignalFunc) dialog_toggle_update,
                      &snvals.use_max_norm);
  gtk_widget_show (toggle);

  /*  Check button #4  */
  toggle = gtk_check_button_new_with_label ("show lines");
  gtk_table_attach (GTK_TABLE (table), toggle, 2, 3, 4, 5,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
                               snvals.showlines);
  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
                      (GtkSignalFunc) dialog_toggle_update,
                      &snvals.showlines);
  gtk_widget_show (toggle);

  /*  Check button #5  */
  toggle = gtk_check_button_new_with_label ("color class");
  gtk_table_attach (GTK_TABLE (table), toggle, 2, 3, 5, 6,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
                               snvals.color_class);
  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
                      (GtkSignalFunc) dialog_toggle_update,
                      &snvals.color_class);
  gtk_widget_show (toggle);

  /*  Check button #6  */
  toggle = gtk_check_button_new_with_label ("color norm");
  gtk_table_attach (GTK_TABLE (table), toggle, 2, 3, 6, 7,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
                               snvals.color_norm);
  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
                      (GtkSignalFunc) dialog_toggle_update,
                      &snvals.color_norm);
  gtk_widget_show (toggle);

  /*  Check button #7  */
  toggle = gtk_check_button_new_with_label ("show points");
  gtk_table_attach (GTK_TABLE (table), toggle, 2, 3, 7, 8,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
                               snvals.showpoints);
  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
                      (GtkSignalFunc) dialog_toggle_update,
                      &snvals.showpoints);
  gtk_widget_show (toggle);

  /*  Check button #8  */
  toggle = gtk_check_button_new_with_label ("use grid");
  gtk_table_attach (GTK_TABLE (table), toggle, 2, 3, 8, 9,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
                               snvals.use_grid);
  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
                      (GtkSignalFunc) dialog_toggle_update,
                      &snvals.use_grid);
  gtk_widget_show (toggle);

  /*  Button #1  */
  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) dialog_ok_callback, dlg);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
                      TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  /*  Button #2  */
  button = gtk_button_new_with_label ("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
                             (GtkSignalFunc) gtk_widget_destroy,
                             GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
                      TRUE, TRUE, 0);
  gtk_widget_show (button);

  gtk_widget_show (table);
  gtk_widget_show (dlg);

  gtk_main ();
  gdk_flush ();

  g_free(temp_ptr);    /* caution: argv might have changed */

  return snint.run;
}  


static void
dialog_close_callback (GtkWidget *widget, gpointer data)
{
  gtk_main_quit ();
}


static void
dialog_toggle_update (GtkWidget *widget, gpointer data)
{
  int *toggle_val;

  toggle_val = (int *) data;

  if (GTK_TOGGLE_BUTTON (widget)->active)
    *toggle_val = TRUE;
  else
    *toggle_val = FALSE;
}


static void
dialog_float_entry_callback (GtkWidget *widget, gpointer data)
{
  gdouble *text_val;

  text_val = (gdouble *) data;

  *text_val = atof (gtk_entry_get_text (GTK_ENTRY (widget)));
}


static void
dialog_int_entry_callback (GtkWidget *widget, gpointer data)
{
  gint *text_val;

  text_val = (gint *) data;

  *text_val = atoi (gtk_entry_get_text (GTK_ENTRY (widget)));
}


static void
dialog_ok_callback (GtkWidget *widget, gpointer data)
{
  snint.run = TRUE;
  gtk_widget_destroy (GTK_WIDGET (data));
}

static void 
dialog_sprint_float(char *buffer, void *pdata)
{
  sprintf(buffer, "%f", *((gdouble*)pdata));
}


static void 
dialog_sprint_int(char *buffer, void *pdata)
{
  sprintf(buffer, "%d", *((gint*)pdata));
}


